# region [Imports]

# * Standard Library Imports -->
import os

# * Gid Imports -->
import gidlogger as glog

# * Local Imports -->
from gidconfig.utility.functions import pathmaker
from gidconfig.experimental.experimental_ini import GidAttConfigIni

# endregion[Imports]

__updated__ = '2020-11-14 01:44:46'

# region [AppUserData]

# endregion [AppUserData]

# region [Logging]

log = glog.aux_library_logger(__name__)
log.info(glog.imported(__name__))

# endregion[Logging]

# region [Constants]

# endregion[Constants]


class ConfigStrategos:
    config_instances = {}
    ini_class = GidAttConfigIni
    json_class = None

    @classmethod
    def borrow_config(cls, config_file):
        config_file = pathmaker(config_file)
        _out = cls.config_instances.get(config_file, None)
        if _out is None:
            _out = cls._create_new_config_instance(config_file)
            cls.config_instances[config_file] = _out
        return _out

    @classmethod
    def _create_new_config_instance(cls, config_file):
        _extension = os.path.basename(config_file).split('.')[-1]
        if _extension == 'ini':
            _cfg = cls._new_ini_config(config_file)
        else:
            return None
        return _cfg

    @classmethod
    def _new_ini_config(cls, config_file):
        _new = cls.ini_class(config_file)
        _new.load()
        return _new

    @classmethod
    def _new_json_config(cls, config_file):
        pass

# region[Main_Exec]


if __name__ == '__main__':
    pass

# endregion[Main_Exec]



# region [Imports]

# * Standard Library Imports -->
import re
from abc import ABC, abstractmethod

# * Gid Imports -->
import gidlogger as glog

# endregion[Imports]

__updated__ = '2020-11-14 15:58:38'

# region [AppUserData]

# endregion [AppUserData]

# region [Logging]

log = glog.aux_library_logger(__name__)
log.info(glog.imported(__name__))

# endregion[Logging]

# region [Constants]

# endregion[Constants]


class GidAttConfigAbstract(ABC):
    def __init__(self, config_file):
        self.config_file = config_file
        self.timedelta_template = "{negative}days: {days}, hours: {hours}, minutes: {minutes}, seconds: {seconds}"
        self.timedelta_regex = re.compile(r"^days:\s*?(?P<days>\d*?),\s*?hours:\s*?(?P<hours>\d*?),\s*?minutes:\s*?(?P<minutes>\d*?),\s*?seconds:\s*?(?P<seconds>\d*?)$")
        self.typus_data = {}
        self.added_attributes = []

    @abstractmethod
    def load(self):
        ...

    @abstractmethod
    def save(self):
        ...

    @abstractmethod
    def new_section(self, section_name, **options):
        ...

    @abstractmethod
    def _edit_dict_attribute(self, section, **kwargs):
        ...

    # region[Main_Exec]


if __name__ == '__main__':
    pass

# endregion[Main_Exec]



# region [Imports]

# * Standard Library Imports -->

# * Gid Imports -->
import gidlogger as glog

# * Local Imports -->
from gidconfig.standard.classes import ConfigHandler
from gidconfig.utility.functions import pathmaker

# endregion [Imports]


# region [Logging]

log = glog.aux_library_logger(__name__)
log.info(glog.imported(__name__))

# endregion [Logging]

# region [Factories]


class ConfigRental:
    config_instances = {}
    appdata = None

    @classmethod
    def set_appdata(cls, appdata_object):
        cls.appdata = appdata_object

    @classmethod
    def get_config(cls, file_name, cfg_folder=None, allow_no_value=True, ** attribute_edits):
        if cls.appdata is None and cfg_folder is None:
            raise FileExistsError('appdata has not been set')
        _folder = cls.appdata['config'] if cfg_folder is None else pathmaker(cfg_folder)
        _file = pathmaker(_folder, file_name)

        _out_cfg = cls.config_instances.get(_file, None)
        if _out_cfg is None:
            _out_cfg = ConfigHandler(config_file=_file, allow_no_value=allow_no_value, **attribute_edits)
            cls.config_instances[_file] = _out_cfg
        return _out_cfg


# endregion [Factories]

# region [Main_Exec]
if __name__ == '__main__':
    pass


# endregion [Main_Exec]



# region [Imports]


# * Standard Library Imports -->
import os
import json

# * Gid Imports -->
import gidlogger as glog

# endregion[Imports]

__updated__ = '2020-11-15 11:24:09'


# region [Logging]

log = glog.aux_library_logger(__name__)
log.info(glog.imported(__name__))

# endregion[Logging]

# region [Constants]

# endregion[Constants]


def pathmaker(first_segment, *in_path_segments, rev=False):
    """
    Normalizes input path or path fragments, replaces '\\\\' with '/' and combines fragments.

    Parameters
    ----------
    first_segment : str
        first path segment, if it is 'cwd' gets replaced by 'os.getcwd()'
    rev : bool, optional
        If 'True' reverts path back to Windows default, by default None

    Returns
    -------
    str
        New path from segments and normalized.
    """
    _first = os.getcwd() if first_segment == 'cwd' else first_segment
    _path = os.path.join(_first, *in_path_segments)
    _path = _path.replace('\\\\', '/')
    _path = _path.replace('\\', '/')
    if rev is True:
        _path = _path.replace('/', '\\')

    return _path.strip()


def writebin(in_file, in_data):
    """
    Writes a string to binary.

    Parameters
    ----------
    in_file : str
        The target file path
    in_data : str
        The data to write
    """
    with open(in_file, 'wb') as outbinfile:
        outbinfile.write(in_data)


def writeit(in_file, in_data, append=False, in_encoding='utf-8', in_errors=None):
    """
    Writes to a file.

    Parameters
    ----------
    in_file : str
        The target file path
    in_data : str
        The data to write
    append : bool, optional
        If True appends the data to the file, by default False
    in_encoding : str, optional
        Sets the encoding, by default 'utf-8'
    """
    _write_type = 'w' if append is False else 'a'
    with open(in_file, _write_type, encoding=in_encoding, errors=in_errors,) as _wfile:
        _wfile.write(in_data)


def appendwriteit(in_file, in_data, in_encoding='utf-8'):
    with open(in_file, 'a', encoding=in_encoding) as appendwrite_file:
        appendwrite_file.write(in_data)


def readbin(in_file):
    """
    Reads a binary file.

    Parameters
    ----------
    in_file : str
        A file path

    Returns
    -------
    str
        the decoded file as string
    """
    with open(pathmaker(in_file), 'rb') as binaryfile:
        return binaryfile.read()


def readit(in_file, per_lines=False, in_encoding='utf-8', in_errors=None):
    """
    Reads a file.

    Parameters
    ----------
    in_file : str
        A file path
    per_lines : bool, optional
        If True, returns a list of all lines, by default False
    in_encoding : str, optional
        Sets the encoding, by default 'utf-8'
    in_errors : str, optional
        How to handle encoding errors, either 'strict' or 'ignore', by default 'strict'

    Returns
    -------
    str/list
        the read in file as string or list (if per_lines is True)
    """
    with open(in_file, 'r', encoding=in_encoding, errors=in_errors) as _rfile:
        _content = _rfile.read()
    if per_lines is True:
        _content = _content.splitlines()

    return _content


def linereadit(in_file, in_encoding='utf-8', in_errors='strict'):
    with open(in_file, 'r', encoding=in_encoding, errors=in_errors) as lineread_file:
        _out = lineread_file.read().splitlines()
    return _out


def clearit(in_file):
    """
    Deletes the contents of a file.

    Parameters
    ----------
    in_file : str
        The target file path
    """
    with open(in_file, 'w') as file_to_clear:
        file_to_clear.write('')
    log.debug(f"contents of file '{in_file}' was cleared")


def loadjson(in_file):
    with open(in_file, 'r') as jsonfile:
        _out = json.load(jsonfile)
    return _out


def writejson(in_object, in_file, sort_keys=True, indent=0):
    with open(in_file, 'w') as jsonoutfile:
        json.dump(in_object, jsonoutfile, sort_keys=sort_keys, indent=indent)


# region[Main_Exec]

if __name__ == '__main__':
    pass

# endregion[Main_Exec]




# region [Imports]

# * Standard Library Imports -->
import os
import configparser
from typing import Union, Callable, List, Set, Tuple, Iterable, Any
from datetime import datetime, timedelta
from pprint import pprint
import re
import asyncio
from hashlib import md5, sha256, sha512, blake2b, blake2s
# * Third Party Imports -->
from fuzzywuzzy import fuzz

# * Gid Imports -->
import gidlogger as glog
import logging
# * Local Imports -->
from gidconfig.data.enums import Get, Fallback
from gidconfig.utility.functions import readit
from gidconfig.utility.decorators import debug_timing_print, debug_timing_log
# endregion [Imports]


# region [Logging]

log = glog.aux_library_logger(__name__)
log.info(glog.imported(__name__))
# endregion [Logging]


class ConfigHandler(configparser.ConfigParser):
    allowed_default_settings = ['defaults',
                                'dict_type',
                                'allow_no_value',
                                'delimiters',
                                'comment_prefixes',
                                'inline_comment_prefixes',
                                'strict',
                                'empty_lines_in_values',
                                'default_section',
                                'interpolation',
                                'converters',
                                'auto_read',
                                'auto_save']

    default_settings = {"auto_read": True,
                        "auto_save": True,
                        "allow_no_value": True}

    def __init__(self, config_file=None, auto_read=True, auto_save=True, allow_no_value=True, **kwargs):
        super().__init__(**kwargs, allow_no_value=allow_no_value)
        self.config_file = '' if config_file is None else config_file
        self.auto_read = auto_read
        self.auto_save = auto_save
        self._method_select = {Get.basic: self.get, Get.boolean: self.getboolean, Get.int: self.getint, Get.list: self.getlist, Get.path: self.get_path, Get.datetime: self.get_datetime}
        self.annotation_replacements = {'[DEFAULT]': 'Options in this section are used if those options are not set in a Section'}

        if self.auto_read is True:
            self.read(self.config_file)

    @classmethod
    def set_default_setting(cls, key, value):
        if key not in cls.allowed_default_settings:
            raise KeyError(f'"{key}" is not an allowed default setting')
        cls.default_settings[key] = value

    def getlist(self, section, option, delimiter=',', as_set=False, casefold_items=False, fallback_option: str = None):
        _raw = self.get(section, option, fallback=self.get(section, fallback_option) if fallback_option is not None else fallback_option).strip()
        if _raw.endswith(delimiter):
            _raw = _raw.rstrip(delimiter)
        if _raw.startswith(delimiter):
            _raw = _raw.lstrip(delimiter).strip()
        _out = _raw.replace(delimiter + ' ', delimiter).split(delimiter)
        _out = list(map(lambda x: x.strip(), _out))
        if casefold_items is True:
            _out = list(map(lambda x: x.casefold(), _out))
        if as_set is True:
            _out = set(_out)
        return _out

    def list_from_keys_only(self, section, as_set=True):
        _result = self.options(section)
        _out = []
        for line in _result:
            if line != '':
                _out.append(line)
        if as_set is True:
            _out = set(_out)
        return _out

    def get_path(self, section, option, cwd_symbol='+cwd+'):
        _raw_path = self.get(section, option)
        if cwd_symbol in _raw_path:
            _out = _raw_path.replace(cwd_symbol, os.getcwd()).replace('\\', '/')
        elif '+userdata+' in _raw_path:
            _out = _raw_path.replace('+userdata+', os.getenv('APPDATA')).replace('\\', '/')
        elif _raw_path == 'notset':
            _out = None
        else:
            _out = os.path.join(_raw_path).replace('\\', '/')
        return _out

    def _best_fuzzymatch(self, in_term, in_targets: Union[list, set, frozenset, tuple, dict]):
        # Todo: replace with process.extractOne() from fuzzywuzzy!
        _rating_list = []
        for _target in in_targets:
            _rating_list.append((_target, fuzz.ratio(in_term, _target)))
        _rating_list.sort(key=lambda x: x[1], reverse=True)
        log.debug("with a fuzzymatch, the term '%s' was best matched to '%s' with and Levenstein-distance of %s", in_term, _rating_list[0][0], _rating_list[0][1])
        return _rating_list[0][0]

    def get_timedelta(self, section, option, amount_seperator=' ', delta_seperator=','):
        _raw_timedelta = self.get(section, option)
        if _raw_timedelta != 'notset':
            _raw_timedelta_list = _raw_timedelta.split(delta_seperator)
            _arg_dict = {'days': 0,
                         'seconds': 0,
                         'microseconds': 0,
                         'milliseconds': 0,
                         'minutes': 0,
                         'hours': 0,
                         'weeks': 0}
            for raw_delta_data in _raw_timedelta_list:
                _amount, _typus = raw_delta_data.strip().split(amount_seperator)
                _key = self._best_fuzzymatch(_typus, _arg_dict)
                _arg_dict[_key] = float(_amount) if '.' in _amount else int(_amount)
            return timedelta(**_arg_dict)

    def get_datetime(self, section, option, dtformat=None):
        _date_time_string = self.get(section, option)
        if _date_time_string == "notset":
            return None
        _dtformat = '%Y-%m-%d %H:%M:%S' if dtformat is None else dtformat
        return datetime.strptime(_date_time_string, _dtformat).astimezone()

    def set_datetime(self, section, option, datetime_object, dtformat=None):
        _dtformat = '%Y-%m-%d %H:%M:%S' if dtformat is None else format
        self.set(section, option, datetime_object.strftime(_dtformat))
        if self.auto_save is True:
            self.save()

    def set(self, section, option, value):
        if isinstance(value, (list, set, tuple)):
            value = ', '.join(list(map(str, value)))
        if not isinstance(value, str):
            value = str(value)
        super().set(section, option, value)
        if self.auto_save is True:
            self.save()

    def enum_get(self, section: str, option: str, typus: Get = Get.basic):
        return self._method_select.get(typus, self.get)(section, option)

    def save(self):
        with open(self.config_file, 'w') as confile:
            self.write(confile)
        self.read()

    def read(self, filename=None):
        _configfile = self.config_file if filename is None else filename
        super().read(_configfile)

    def set_annotation_replacement(self, replacements: dict):
        for key, value in replacements.items():
            self.annotation_replacements[key] = value

    def get_annotated_config(self):
        content = readit(self.config_file)
        for target, annotation in self.annotation_replacements.items():
            replacement = f";; {annotation}\n{target}"
            content = content.replace(target, replacement)
        return content

    def options(self, section: str) -> List[str]:

        return [option for option in super().options(section) if option not in self.defaults()]

    @classmethod
    def from_defaults(cls, config_file):
        return cls(config_file, **cls.default_settings)

    def __contains__(self, item: object):
        return item in self.sections()

    def __str__(self):
        return os.path.splitext(os.path.basename(self.config_file))[0].upper()


class SingleAccessConfigHandler(ConfigHandler):

    allowed_default_settings = ['defaults',
                                'dict_type',
                                'delimiters',
                                'inline_comment_prefixes',
                                'strict',
                                'empty_lines_in_values',
                                'default_section',
                                'interpolation',
                                'converters',
                                'auto_read',
                                'auto_save',
                                'list_delimiter',
                                'comment_marker',
                                'top_comment',
                                'read_before_retrieve']

    default_settings = {"auto_read": True,
                        "auto_save": True,
                        "allow_no_value": True,
                        "list_delimiter": ',',
                        "comment_marker": '#',
                        "comment_prefixes": '~',
                        "read_before_retrieve": False}

    bool_true_values = {'yes', 'true', '+', 'y', 'on', 'enabled', 'positive'}
    bool_false_values = {'no', 'false', '-', 'n', 'off', 'disabled', 'negative'}

    def __init__(self, config_file=None, auto_read=True, auto_save=True, allow_no_value=True, list_delimiter=',', read_before_retrieve=False, comment_marker='#', top_comment: Union[str, Iterable] = None, comment_prefixes='~', ** kwargs):
        self.section_comments = {}
        self.read_before_retrieve = read_before_retrieve
        self.section_header_size = 150
        self.section_header_chars = {'top': '▲', 'bottom': '▼', 'middle': '▌'}
        self.comment_marker = comment_marker
        self.top_comment_regex = re.compile(r"^[^\[\]]*?(?=\[)", re.DOTALL)
        self._top_comment = self._validate_top_comment(top_comment) if top_comment is not None else top_comment
        super().__init__(config_file=config_file, auto_read=auto_read, auto_save=auto_save, allow_no_value=allow_no_value, comment_prefixes=comment_prefixes, ** kwargs)
        self.list_delimiter = list_delimiter
        self.file_hash_when_loaded = ""

        self.typus_table = {Any: self._auto_convert_value,
                            str: str,
                            int: int,
                            float: float,
                            list: self._as_list,
                            List: self._as_list,
                            List[str]: self._as_list,
                            List[int]: self._as_list_int,
                            List[float]: self._as_list_float,
                            List[bool]: self._as_list_bool,
                            set: self._as_set,
                            Set: self._as_set,
                            Set[str]: self._as_set,
                            Set[int]: self._as_set_int,
                            Set[float]: self._as_set_float,
                            Set[bool]: self._as_set_bool,
                            tuple: self._as_tuple,
                            Tuple: self._as_tuple,
                            Tuple[str]: self._as_tuple,
                            Tuple[int]: self._as_tuple_int,
                            Tuple[float]: self._as_tuple_float,
                            Tuple[bool]: self._as_tuple_bool,
                            bool: self._as_bool}

    @debug_timing_log(log)
    def get_config_file_hash(self, in_file=None):
        _file = self.config_file if in_file is None else in_file
        with open(_file, 'rb') as f:
            return md5(f.read()).hexdigest()

    @property
    def top_comment(self):
        return self._top_comment

    @top_comment.setter
    def top_comment(self, value: Union[Iterable, str]):
        self._top_comment = self._validate_top_comment(value)

    def _validate_top_comment(self, top_comment):
        if isinstance(top_comment, str):
            top_comment = top_comment.splitlines()
        _out = []
        for line in top_comment:
            if not line.startswith(self.comment_marker):
                line = f"{self.comment_marker} {line.strip()}"
            _out.append(line)
        return '\n'.join(_out)

    @debug_timing_log(log)
    def retrieve(self, section, option, typus=Any, *, fallback_section: str = None, fallback_option: str = None, direct_fallback=Fallback.NULL, mod_func: Callable = None):
        if self.read_before_retrieve is True or self.file_hash_when_loaded != self.get_config_file_hash():
            log.debug("config file %s has changed content", os.path.basename(self.config_file))
            self.read()

        raw_data = self.get(section=section, option=option, fallback=None)
        raw_data = raw_data if raw_data not in ['', None] else self.defaults().get(option, None)

        if raw_data in [None, '']:
            if direct_fallback is not Fallback.NULL:
                return direct_fallback
            if fallback_option is not None:
                fallback_section = section if fallback_section is None else fallback_section
                raw_data = self.get(fallback_section, fallback_option, fallback=None)
                raw_data = raw_data if raw_data not in ['', None] else self.defaults().get(fallback_option)
            else:
                if self.has_section(section) is False:
                    raise configparser.NoSectionError(section)
                if self.has_option(section, option) is False:
                    raise configparser.NoOptionError(option, section)

        data = self.typus_table.get(typus)(raw_data)
        if mod_func is not None:
            if type(data) in [list, tuple, set]:
                data = type(data)(map(mod_func, data))
            else:
                data = mod_func(data)
        return data

    def _as_list(self, raw_data):
        raw_data = raw_data.strip(self.list_delimiter).strip()

        list_data = raw_data.split(self.list_delimiter)
        return list(map(lambda x: x.strip(), list_data))

    def _as_list_int(self, raw_data):
        list_data = self._as_list(raw_data)
        return list(map(int, list_data))

    def _as_list_float(self, raw_data):
        list_data = self._as_list(raw_data)
        return list(map(float, list_data))

    def _as_list_bool(self, raw_data):
        list_data = self._as_list(raw_data)
        return list(map(self._as_bool, list_data))

    def _as_set(self, raw_data):
        return set(self._as_list(raw_data))

    def _as_set_int(self, raw_data):
        return set(self._as_list_int(raw_data))

    def _as_set_float(self, raw_data):
        return set(self._as_list_float(raw_data))

    def _as_set_bool(self, raw_data):
        return set(self._as_list_bool(raw_data))

    def _as_tuple(self, raw_data):
        return tuple(self._as_list(raw_data))

    def _as_tuple_int(self, raw_data):
        return tuple(self._as_list_int(raw_data))

    def _as_tuple_float(self, raw_data):
        return tuple(self._as_list_float(raw_data))

    def _as_tuple_bool(self, raw_data):
        return tuple(self._as_list_bool(raw_data))

    def _as_bool(self, raw_data):
        raw_data = raw_data.casefold()
        if raw_data in self.bool_true_values:
            return True
        if raw_data in self.bool_false_values:
            return False
        else:
            raise ValueError(f"'{raw_data}' can not converted to a bool value")

    def _make_section_header(self, section):
        section = f" {section.upper()} "
        top = f'{self.comment_marker} {self.section_header_chars.get("top")*self.section_header_size}'
        middle = f'{self.comment_marker} {section.center(self.section_header_size, self.section_header_chars.get("middle"))}'
        bottom = f'{self.comment_marker} {self.section_header_chars.get("bottom")*self.section_header_size}'
        return '\n'.join(['\n', top, middle, bottom, '\n'])

    def _reset_section_comments(self, filename=None):
        filename = self.config_file if filename is None else filename
        _new_content = []
        with open(filename, 'r') as f:
            old_content = f.read()
        content_lines = old_content.splitlines()
        for index, line in enumerate(content_lines):
            for key, value in self.section_comments.items():
                if line == f'[{key}]':
                    line = self._make_section_header(key) + '\n'.join(value + [f'[{key}]'])
            if not line.startswith(self.comment_marker) and line != '' and not line.startswith('\t') and not content_lines[index + 1].startswith('\t'):
                line = line + '\n'
            _new_content.append(line)
        with open(filename, 'w') as f:
            if self.top_comment is not None:
                f.write(self._make_section_header('instructions') + '\n')
                f.write(self.top_comment + '\n\n' + f'# {"-"*self.section_header_size}\n\n')
            f.write('\n'.join(_new_content))

    def append(self, section, option, value):
        if not self.has_section(section):
            raise configparser.NoSectionError(section)
        if not self.has_option(section, option):
            raise configparser.NoOptionError(option, section)
        if isinstance(value, (list, set, tuple)):
            value = ', '.join(value)
        if not isinstance(value, str):
            value = str(value)

        data = self.retrieve(section, option, typus=str)
        data += ', ' + value
        self.set(section, option, data)

    @debug_timing_log(log)
    def save(self, filename=None):
        filename = self.config_file if filename is None else filename
        with open(filename, 'w') as f:
            self.write(f)
        self._reset_section_comments(filename)

    async def async_save(self, filename=None):
        await asyncio.to_thread(self.save, filename=filename)

    @debug_timing_log(log)
    def read(self, filename=None):
        _configfile = self.config_file if filename is None else filename
        if filename is not None:
            self.config_file = filename
        self.file_hash_when_loaded = self.get_config_file_hash()
        _cleaned_content = []
        _buffered_comment_lines = []
        _current_section = None
        with open(_configfile, 'r') as f:
            content_lines = f.read().splitlines()
        for line in reversed(content_lines):

            if line.startswith('['):
                _current_section = line.strip('[]')

            elif line.startswith(f"{self.comment_marker} {self.section_header_chars.get('top')}") or line.startswith(f"{self.comment_marker} {self.section_header_chars.get('middle')}") or line.startswith(f"{self.comment_marker} {self.section_header_chars.get('bottom')}"):
                line = None
            elif _current_section is not None and line.startswith(self.comment_marker):
                _buffered_comment_lines.append(line)
                line = None
            elif _current_section is not None and not line.startswith(self.comment_marker):
                self.section_comments[_current_section] = list(reversed(_buffered_comment_lines))
                _current_section = None
                _buffered_comment_lines = []

            if line is not None:
                _cleaned_content.append(line)
        if _current_section is not None:
            self.section_comments[_current_section] = list(reversed(_buffered_comment_lines))

        new_content = '\n'.join(reversed(_cleaned_content))
        if self.top_comment is None:
            match = self.top_comment_regex.match(new_content)
            if match:
                self.top_comment = match.group().replace(self._make_section_header('instructions') + '\n', '').replace(f'# {"-"*self.section_header_size}\n\n', '').strip('\n')
        self.clear()
        super().read_string(self.top_comment_regex.sub('', new_content))

    def options(self, section: str) -> List[str]:
        options = super().options(section)
        return [option for option in options if self.comment_marker not in option]

    def _list_auto_convert_value(self, value: str):
        _out = []
        for item in value.split(self.list_delimiter):
            item = item.strip()
            _out.append(self._auto_convert_value(item))
        return _out

    def _auto_convert_value(self, value: str):
        if value is None:
            return None
        value = value.casefold()
        if self.list_delimiter in value:
            return self._list_auto_convert_value(value)
        if value.isdigit() is True:
            return int(value)
        if '.' in value and value.replace('.', '', 1).isdigit():
            return float(value)
        if value in self.bool_true_values:
            return True
        if value in self.bool_false_values:
            return False
        return value

    @debug_timing_log(log)
    def to_dict(self):
        _out = {'default': {}}
        for default_option, default_value in self.defaults().items():
            if self.comment_marker not in default_option:
                _out['default'][default_option] = self._auto_convert_value(default_value)
        for section in self.sections():
            _out[section] = {}
            for option in self.options(section):
                value = self.retrieve(section, option, typus=str, direct_fallback=None)
                _out[section][option] = self._auto_convert_value(value)
        return _out

    async def async_to_dict(self):
        return await asyncio.to_thread(self.to_dict)


if __name__ == '__main__':
    pass



# region [Imports]

# * Standard Library Imports -->
from datetime import datetime, timedelta
from functools import partial
from collections import namedtuple
from configparser import ConfigParser
from pprint import pprint
# * Gid Imports -->
import gidlogger as glog

# * Local Imports -->
from gidconfig.experimental.experimental_abstract import GidAttConfigAbstract

# endregion[Imports]

__updated__ = '2020-11-15 04:21:50'

# region [AppUserData]

# endregion [AppUserData]

# region [Logging]

log = glog.aux_library_logger(__name__)
log.info(glog.imported(__name__))

# endregion[Logging]

# region [Constants]

# endregion[Constants]


class GidAttConfigIni(GidAttConfigAbstract):
    forbidden_section_chars = [' ', '.', '-']
    typus_tuple = namedtuple('TypusTuple', ['typus', 'prefix', 'transformer'])

    def __init__(self, config_file):
        super().__init__(config_file)
        self.configparser = ConfigParser(allow_no_value=True, interpolation=None)
        self.list_seperator = ','
        self.typus_data = {'list': self.typus_tuple(list, '-LIST-', self._list_transformer),
                           'set': self.typus_tuple(set, '-SET-', self._set_transformer),
                           'tuple': self.typus_tuple(tuple, '-TUPLE-', self._tuple_transformer),
                           'float': self.typus_tuple(float, '-FLOAT-', self._float_transformer),
                           'datetime': self.typus_tuple(datetime, '-DATETIME-', self._datetime_transform),
                           'timedelta': self.typus_tuple(timedelta, '-TIMEDELTA-', self._timedelta_transform)}
        self.stored_comments = {}

    # def _store_comments(self):
    #     self.stored_comments = {}
    #     _section = ''
    #     with open(self.config_file, 'r') as conf_f:
    #         content_lines = conf_f.read().splitlines()
    #     for index, line in enumerate(content_lines):
    #         if line.startswith('['):
    #             _section = line
    #         if line.startswith('#'):
    #             if _section not in self.stored_comments:
    #                 self.stored_comments[_section] = []
    #             self.stored_comments[_section].append((content_lines[index + 1], line))

    def load(self):

        for added_attribute in self.added_attributes:
            delattr(self, added_attribute)
        # self._store_comments()
        self.added_attributes = []
        self.configparser.read(self.config_file)
        for section in self.configparser.sections():
            if any(forbidden_char in section for forbidden_char in self.forbidden_section_chars):
                raise AttributeError(f"forbidden character in section name '{section}'")
            setattr(self, section, {})
            self.added_attributes.append(section)
            setattr(self, 'set_' + section, partial(self._edit_dict_attribute, section))
            self.added_attributes.append('set_' + section)
            for option in self.configparser.options(section):
                _value = self.configparser.get(section, option)
                _value = self._check_convert_int(_value)
                _value = self._check_convert_boolean(_value)
                _value = self._check_convert_float(_value)
                for _, typus_item in self.typus_data.items():
                    try:
                        if typus_item.prefix in _value:
                            _value = _value.replace(typus_item.prefix, '')
                            _value = typus_item.transformer(_value)
                    except TypeError:
                        pass

                getattr(self, section)[option] = _value

    def _list_transformer(self, in_value, direction='load'):
        if direction == "load":
            _out = in_value.strip().split(self.list_seperator)
            _out = list(map(self._strip_it, _out))
            _out = list(map(self._check_convert_int, _out))
            _out = list(map(self._check_convert_boolean, _out))
            _out = list(map(self._check_convert_float, _out))
        elif direction == 'save':
            _item_string = ', '.join(map(str, in_value))
            _out = self.typus_data['list'].prefix + ' ' + _item_string
        return _out

    @ staticmethod
    def _check_convert_boolean(item):
        _bool_true_values = ['yes', 'true', 'on']
        _bool_false_values = ['no', 'false', 'off']
        try:
            if any(btrue.casefold() == item.casefold() for btrue in _bool_true_values):
                return True
            elif any(bfalse.casefold() == item.casefold() for bfalse in _bool_false_values):
                return False
            else:
                return item
        except AttributeError:
            return item

    @ staticmethod
    def _strip_it(item):
        return item.strip()

    @ staticmethod
    def _check_convert_int(item):
        try:
            _out = int(item)
        except ValueError:
            _out = item
        except TypeError:
            _out = item
        return _out

    @ staticmethod
    def _check_convert_float(item):
        if isinstance(item, bool):
            return item
        try:
            _out = float(item)
        except ValueError:
            _out = item
        except TypeError:
            _out = item
        return _out

    @ staticmethod
    def _set_transformer(item, direction='load'):
        pass

    @ staticmethod
    def _float_transformer(item, direction='load'):
        pass

    @ staticmethod
    def _tuple_transformer(item, direction='load'):
        pass

    def _datetime_transform(self, item, direction='load'):
        if direction == 'load':
            return datetime.strptime(item.strip(), "%Y-%m-%dT%H:%M:%S")
        elif direction == 'save':
            return self.typus_data['datetime'].prefix + ' ' + item.isoformat(timespec='seconds')

    def _timedelta_transform(self, item, direction='load'):
        if direction == 'load':
            _neg = ''
            if 'negative' in item:
                _neg = '-'
            _match = self.timedelta_regex.search(item.replace('negative', '').strip())

            _out = {key: int(_neg + value) for key, value in _match.groupdict().items()}

            return timedelta(**_out)

        elif direction == 'save':
            _total_seconds = item.total_seconds()
            _neg = 'negative ' if str(_total_seconds).startswith('-') else ''
            _total_seconds = abs(_total_seconds)
            _days, remainder = divmod(_total_seconds, 60 * 60 * 24)
            _hours, remainder = divmod(remainder, 60 * 60)
            _minutes, _seconds = divmod(remainder, 60)
            _out = self.typus_data['timedelta'].prefix + ' '
            return _out + self.timedelta_template.format(negative=_neg, days=round(_days), hours=round(_hours), minutes=round(_minutes), seconds=round(_seconds))

    def save(self):
        with open(self.config_file, 'w') as confile:
            self.configparser.write(confile)
        self.load()

    def _edit_dict_attribute(self, section, **kwargs):
        for key, value in kwargs.items():
            for _, typus_item in self.typus_data.items():
                if isinstance(value, typus_item.typus):
                    value = typus_item.transformer(value, direction='save')
            self.configparser.set(section, key, str(value))
        self.save()

    def add_comment(self, section, option, comment):
        orig_value = self.configparser.get(section, option)
        self.configparser.remove_option(section, option)
        self.configparser.set(section, ';' + comment, '')
        self.configparser.set(section, option, orig_value)

    def new_section(self, section_name, **options):
        self.configparser.add_section(section_name)
        self._edit_dict_attribute(section_name, **options)
        self.save()

    def register_new_typus(self, name, compare_type, prefix, transformer):
        self.typus_data[name] = self.typus_tuple(compare_type, prefix, transformer)
        self.save()

# region[Main_Exec]


if __name__ == '__main__':
    pass
# endregion[Main_Exec]



# region [Imports]


# * Standard Library Imports -->
from enum import Enum, Flag, auto

# * Gid Imports -->
import gidlogger as glog

# endregion [Imports]

__updated__ = '2020-11-14 14:54:23'

# region [Logging]

log = glog.aux_library_logger(__name__)
log.info(glog.imported(__name__))

# endregion [Logging]


class Cfg(Flag):
    User = auto()
    Solid = auto()
    Database = auto()
    DefaultFolder = auto()


class Get(Enum):
    basic = auto()
    boolean = auto()
    int = auto()
    list = auto()
    path = auto()
    datetime = auto()


class Fallback(Enum):
    NULL = auto()


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]
